
/*
 *
 * In Engineerese:
 *   This will take a 44 kHz Stereo PCM Array and return a 11 kHz Stereo PCM Array
 *   with corrections for Aliasing using an FIR (Finite Impulse Response) filter.
 *
 * In English:
 *   When a PCM file (Voice File) is converted from 44.1 kHz (CD Freq) to 11.025 kHz
 *   (Standard Archive Quality) a distortion is seen in the sound.  This is most
 *	 noticable in female vocalists and the letter "s".  Using this algorythm to
 *   convert from the original 44.1 kHz to 11.025 kHz will solve this problem.
 *
 * Written By:
 *		Erik H Gawtry and Dennis Pfab		June, 1995
 *
 * Note:
 *		"short" used instead of "int" just in case someone is using UNICODE.
 *
 * Bugs:
 *		Slow as Frozen Molasses
 *
 * Usage Example:
 *
 *		firconvert	fc;
 *		int			size;
 *		char		szBuffer[BUFLEN], szNew[BUFLEN/4];
 *
 *		 // Reset the FIR filter
 *		fc.Reset();
 *
 *		while( (size=_read(fd, szBuffer, BUFLEN)) > 0 )
 *		{
 *			 // Convert the read data
 *			 // Note: (size/2)       for two stereo channels
 *			 //       /sizeof(short) for data item size
 *			size = fc._44_to_11(szBuffer, szNew, (size/2)/sizeof(short));
 *
 *			 // Write the converted data
 *			_write(fd, szNew, size*sizeof(short));
 *		}
 *
 *
 *
 */

#ifndef _FIR
#define	_FIR

#include <memory.h>

typedef unsigned int	UINT;
typedef short _far		*LPSHORT;

 // computed FIR Length
#define	FIR_LEN	97

 // computed FIR Filter coeffecients
const short _nfir[FIR_LEN] = {
			-28, -15, 4, 41, 82, 107, 97, 49,
			-20, -77, -89, -43, 38, 112, 130, 69,
			-43, -149, -179, -103, 49, 196, 245, 150,
			-54, -258, -334, -214, 58, 341, 457, 307,
			-62, -460, -642, -452, 65, 655, 960, 716, 
			-67, -1052, -1670, -1379, 69, 2454, 5119, 7205, 
			7993, // coefficient of tap 49
			7205, 5119, 2454, 69, -1379, -1670, -1052, -67, 
			716, 960, 655, 65, -452, -642, -460, -62, 
			307, 457, 341, 58, -214, -334, -258, -54, 
			150, 245, 196, 49, -103, -179, -149, -43, 
			69, 130, 112, 38, -43, -89, -77, -20, 
			49, 97, 107, 82, 41, 4, -15, -28 };


class firconvert
{
	short	*lArray;	// Last Array
	int		nArrayLen;	// Length of Last Array
public:
	firconvert()
	{
		lArray = new short[(FIR_LEN+1)*2];
		nArrayLen = 0;
	}
	 // Reset the FIR filter for new Stereo Word Array
	void Reset(void) 
	{
		nArrayLen = 0;
	}
	 // Stereo Word Array and Number of Elements for arguments
	UINT _44_to_11(LPSHORT pArrayIn, LPSHORT pArrayOut, UINT nLen)
	{
		 // Get a pointer to the FIR coeffecients
		LPSHORT pCoefStart = (LPSHORT)_nfir;

		 // If a continuation, start from beginning, otherwise use index
		UINT	nCoefLen;
		if( nArrayLen > 0 )
			nCoefLen = 1;
		else
			nCoefLen = (FIR_LEN+1)/2;
		
		 // Set start and end points
		LPSHORT pInEnd = pArrayIn+nLen*2;
		LPSHORT pIn    = pArrayIn+nCoefLen*2-1;
		
		 // Start with current indent
		UINT	nAccLen = nCoefLen;
		
		 // Accumulators for result
		long	nAccL = 0, nAccR = 0;
		UINT	i;
		
		 // Output Array Length
		UINT	nOutLen = 0;
		
		 // Temp array pointers to speed things up
		register LPSHORT pData, pCoef;
		
		 // Loop through entire input array (by 4s for freq. conversion)
		for(UINT p=0; p < nLen; p += 4)
		{
			 // Set the tmp pointers
			pData = pIn;
			pCoef = pCoefStart;
			
			 // Get initial values
			nAccL = (long)(*pCoef) * (*pData--);
			nAccR = (long)(*pCoef) * (*pData--);
			pCoef++;

			 // Do FIR adding			
			for(i=1; i < nAccLen; ++i, pCoef++)
			{
				nAccL += (long)(*pCoef) * (*pData--);
				nAccR += (long)(*pCoef) * (*pData--);
			}
			
			 // If starting and last array values available, use 'em
			if( nArrayLen > 0 && i < FIR_LEN )
			{
				 // Use the last iteration values to complete FIR filter
				pData = lArray+nArrayLen*2-1;
				for(; i < FIR_LEN; ++i, pCoef++)
				{
					nAccL += (long)(*pCoef) * (*pData--);
					nAccR += (long)(*pCoef) * (*pData--);
				}
			}
			
			 // Store results (Note: Right and Left Reversed)
			pArrayOut[nOutLen++] = (int)(nAccR>>16);
			pArrayOut[nOutLen++] = (int)(nAccL>>16);
			
			 // If at end, reduce size of coef array
			if( pIn >= pInEnd )
			{
				 // Decrease by 4 for freq. conversion
				nAccLen-=4;
				pCoefStart+=4;
			}
			 // If at start, increase size of coef array to full size
			else
			{
				 // Increase by 4 for freq. conversion
				if( nAccLen < FIR_LEN ) nAccLen += 4;
				pIn += 8;
			}
		}
		 // Copy array for use in next iteration 
		 //
		nArrayLen = FIR_LEN;
		memcpy(lArray, pInEnd-FIR_LEN*2, FIR_LEN*2*sizeof(short));
		
		 // Return modified length (/4)
		return nOutLen;
	}
};

#endif
